@@ -36,6 +36,8 @@
 #include "Message.h"
 #include "Plugins.h"
 #include "PacketDataStream.h"
+#include "ServerHandler.h"
+#include "VoiceRecorder.h"
 
 // Remember that we cannot use static member classes that are not pointers, as the constructor
 // for AudioOutputRegistrar() might be called before they are initialized, as the constructor
@@ -89,6 +91,10 @@ bool AudioOutputRegistrar::usesOutputDelay() const {
 	return true;
 }
 
+bool AudioOutputRegistrar::canExclusive() const {
+	return false;
+}
+
 AudioOutputUser::AudioOutputUser(const QString name) : qsName(name) {
 	iBufferSize = 0;
 	pfBuffer = NULL;
@@ -97,10 +103,8 @@ AudioOutputUser::AudioOutputUser(const QString name) : qsName(name) {
 }
 
 AudioOutputUser::~AudioOutputUser() {
-	if (pfBuffer)
-		delete [] pfBuffer;
-	if (pfVolume)
-		delete [] pfVolume;
+	delete [] pfBuffer;
+	delete [] pfVolume;
 }
 
 void AudioOutputUser::resizeBuffer(unsigned int newsize) {
@@ -258,10 +262,8 @@ AudioOutputSample::~AudioOutputSample() {
 	if (srs)
 		speex_resampler_destroy(srs);
 
-	if (sfHandle) {
-		delete sfHandle;
-		sfHandle = NULL;
-	}
+	delete sfHandle;
+	sfHandle = NULL;
 }
 
 SoundFile* AudioOutputSample::loadSndfile(const QString &filename) {
@@ -318,7 +320,7 @@ bool AudioOutputSample::needSamples(unsigned int snum) {
 		return true;
 
 	// Calculate the required buffersize to hold the results
-	unsigned int iInputFrames = iroundf(ceilf(static_cast<float>(snum * sfHandle->samplerate()) / static_cast<float>(iOutSampleRate)));
+	unsigned int iInputFrames = static_cast<unsigned int>(ceilf(static_cast<float>(snum * sfHandle->samplerate()) / static_cast<float>(iOutSampleRate)));
 	unsigned int iInputSamples = iInputFrames * sfHandle->channels();
 
 	float *pOut;
@@ -401,7 +403,7 @@ AudioOutputSpeech::AudioOutputSpeech(ClientUser *user, unsigned int freq, Messag
 	else
 		srs = NULL;
 
-	iOutputSize = iroundf(ceilf(static_cast<float>(iFrameSize * freq) / static_cast<float>(srate)));
+	iOutputSize = static_cast<unsigned int>(ceilf(static_cast<float>(iFrameSize * freq) / static_cast<float>(srate)));
 
 	iBufferOffset = iBufferFilled = iLastConsume = 0;
 	bLastAlive = true;
@@ -465,6 +467,7 @@ void AudioOutputSpeech::addFrameToBuffer(const QByteArray &qbaPacket, unsigned i
 		jbp.span = iFrameSize * frames;
 		jbp.timestamp = iFrameSize * iSeq;
 
+#ifdef REPORT_JITTER
 		if (g.s.bUsage && (umtType != MessageHandler::UDPVoiceSpeex) && p && ! p->qsHash.isEmpty() && (p->qlTiming.count() < 3000)) {
 			QMutexLocker qml(& p->qmTiming);
 
@@ -482,6 +485,7 @@ void AudioOutputSpeech::addFrameToBuffer(const QByteArray &qbaPacket, unsigned i
 
 			p->qlTiming.append(jr);
 		}
+#endif
 
 		jitter_buffer_put(jbJitter, &jbp);
 	}
@@ -510,8 +514,9 @@ bool AudioOutputSpeech::needSamples(unsigned int snum) {
 		if (! bLastAlive) {
 			memset(pOut, 0, iFrameSize * sizeof(float));
 		} else {
-			if (p == &LoopUser::lpLoopy)
+			if (p == &LoopUser::lpLoopy) {
 				LoopUser::lpLoopy.fetchFrames();
+			}
 
 			int avail = 0;
 			int ts = jitter_buffer_get_pointer_timestamp(jbJitter);
@@ -599,7 +604,7 @@ bool AudioOutputSpeech::needSamples(unsigned int snum) {
 						}
 					}
 					if (cdDecoder)
-						cCodec->celt_decode_float(cdDecoder, qba.isEmpty() ? NULL : reinterpret_cast<const unsigned char *>(qba.constData()), qba.size(), pOut);
+						cCodec->decode_float(cdDecoder, qba.isEmpty() ? NULL : reinterpret_cast<const unsigned char *>(qba.constData()), qba.size(), pOut);
 					else
 						memset(pOut, 0, sizeof(float) * iFrameSize);
 				} else {
@@ -644,7 +649,7 @@ bool AudioOutputSpeech::needSamples(unsigned int snum) {
 			} else {
 				if (umtType != MessageHandler::UDPVoiceSpeex) {
 					if (cdDecoder)
-						cCodec->celt_decode_float(cdDecoder, NULL, 0, pOut);
+						cCodec->decode_float(cdDecoder, NULL, 0, pOut);
 					else
 						memset(pOut, 0, sizeof(float) * iFrameSize);
 				} else {
@@ -673,21 +678,21 @@ nextframe:
 	}
 
 	if (p) {
-		ClientUser::TalkState ts;
+		Settings::TalkState ts;
 		if (! nextalive)
 			ucFlags = 0xFF;
 		switch (ucFlags) {
 			case 0:
-				ts = ClientUser::Talking;
+				ts = Settings::Talking;
 				break;
 			case 1:
-				ts = ClientUser::TalkingWhisperChannel;
+				ts = Settings::Shouting;
 				break;
 			case 0xFF:
-				ts = ClientUser::TalkingOff;
+				ts = Settings::Passive;
 				break;
 			default:
-				ts = ClientUser::TalkingWhisper;
+				ts = Settings::Whispering;
 				break;
 		}
 		p->setTalking(ts);
@@ -701,7 +706,7 @@ nextframe:
 
 AudioOutput::AudioOutput() {
 	iFrameSize = SAMPLE_RATE / 100;
-	bRunning = false;
+	bRunning = true;
 
 	iChannels = 0;
 	fSpeakers = NULL;
@@ -718,12 +723,9 @@ AudioOutput::~AudioOutput() {
 	wait();
 	wipe();
 
-	if (fSpeakers)
-		delete [] fSpeakers;
-	if (fSpeakerVolume)
-		delete [] fSpeakerVolume;
-	if (bSpeakerPositional)
-		delete [] bSpeakerPositional;
+	delete [] fSpeakers;
+	delete [] fSpeakerVolume;
+	delete [] bSpeakerPositional;
 }
 
 // Here's the theory.
@@ -781,7 +783,7 @@ AudioOutputSample *AudioOutput::playSample(const QString &filename, bool loop) {
 	if (handle == NULL)
 		return NULL;
 
-	while ((iMixerFreq == 0) && bRunning) {
+	while ((iMixerFreq == 0) && isAlive()) {
 #if QT_VERSION >= 0x040500
 		QThread::yieldCurrentThread();
 #endif
@@ -811,14 +813,18 @@ void AudioOutput::addFrameToBuffer(ClientUser *user, const QByteArray &qbaPacket
 		if (aop)
 			removeBuffer(aop);
 
-		while ((iMixerFreq == 0) && bRunning) {}
+		while ((iMixerFreq == 0) && isAlive()) {
+#if QT_VERSION >= 0x040500
+			QThread::yieldCurrentThread();
+#endif
+		}
 
 		if (! iMixerFreq)
 			return;
 
 		qrwlOutputs.lockForWrite();
 		aop = new AudioOutputSpeech(user, iMixerFreq, type);
-		qmOutputs.replace(user,aop);
+		qmOutputs.replace(user, aop);
 	}
 
 	aop->addFrameToBuffer(qbaPacket, iSeq);
@@ -843,12 +849,9 @@ void AudioOutput::removeBuffer(AudioOutputUser *aop) {
 }
 
 void AudioOutput::initializeMixer(const unsigned int *chanmasks, bool forceheadphone) {
-	if (fSpeakers)
-		delete fSpeakers;
-	if (bSpeakerPositional)
-		delete bSpeakerPositional;
-	if (fSpeakerVolume)
-		delete fSpeakerVolume;
+	delete fSpeakers;
+	delete bSpeakerPositional;
+	delete fSpeakerVolume;
 
 	fSpeakers = new float[iChannels * 3];
 	bSpeakerPositional = new bool[iChannels];
@@ -970,16 +973,28 @@ bool AudioOutput::mix(void *outbuff, unsigned int nsamp) {
 	if (g.s.fVolume < 0.01)
 		return false;
 
+	const float adjustFactor = std::pow(10, -18. / 20);
 	const float mul = g.s.fVolume;
 	const unsigned int nchan = iChannels;
+	ServerHandlerPtr sh = g.sh;
+	VoiceRecorderPtr recorder;
+	if (sh)
+		recorder = g.sh->recorder;
 
 	qrwlOutputs.lockForRead();
-	foreach(aop, qmOutputs) {
+	bool needAdjustment = false;
+	QMultiHash<const ClientUser *, AudioOutputUser *>::const_iterator i = qmOutputs.constBegin();
+	while (i != qmOutputs.constEnd()) {
+		AudioOutputUser *aop = i.value();
 		if (! aop->needSamples(nsamp)) {
 			qlDel.append(aop);
 		} else {
 			qlMix.append(aop);
+			// Set a flag if there is a priority speaker
+			if (i.key() && i.key()->bPrioritySpeaker)
+				needAdjustment = true;
 		}
+		++i;
 	}
 
 	if (! qlMix.isEmpty()) {
@@ -992,6 +1007,12 @@ bool AudioOutput::mix(void *outbuff, unsigned int nsamp) {
 
 		memset(output, 0, sizeof(float) * nsamp * iChannels);
 
+		boost::shared_array<float> recbuff;
+		if (recorder) {
+			recbuff = boost::shared_array<float>(new float[nsamp]);
+			memset(recbuff.get(), 0, sizeof(float) * nsamp);
+		}
+
 		for (unsigned int i=0;i<iChannels;++i)
 			svol[i] = mul * fSpeakerVolume[i];
 
@@ -1062,6 +1083,44 @@ bool AudioOutput::mix(void *outbuff, unsigned int nsamp) {
 
 		foreach(aop, qlMix) {
 			const float * RESTRICT pfBuffer = aop->pfBuffer;
+			float volumeAdjustment = 1;
+
+			// We have at least one priority speaker
+			if (needAdjustment) {
+				AudioOutputSpeech *aos = qobject_cast<AudioOutputSpeech *>(aop);
+				// Exclude whispering people
+				if (aos && (aos->p->tsState == Settings::Talking || aos->p->tsState == Settings::Shouting)) {
+					// Adjust all non-priority speakers
+					if (!aos->p->bPrioritySpeaker)
+						volumeAdjustment = adjustFactor;
+				}
+			}
+
+			if (recorder) {
+				AudioOutputSpeech *aos = qobject_cast<AudioOutputSpeech *>(aop);
+
+				if (aos) {
+					for (unsigned int i = 0; i < nsamp; ++i) {
+						recbuff[i] += pfBuffer[i] * volumeAdjustment;
+					}
+
+					if (!recorder->getMixDown()) {
+						if (aos) {
+							recorder->addBuffer(aos->p, recbuff, nsamp);
+						} else {
+							// this should be unreachable
+							Q_ASSERT(false);
+						}
+						recbuff = boost::shared_array<float>(new float[nsamp]);
+						memset(recbuff.get(), 0, sizeof(float) * nsamp);
+					}
+
+					// Don't add the local audio to the real output
+					if (qobject_cast<RecordUser *>(aos->p)) {
+						continue;
+					}
+				}
+			}
 
 			if (validListener && ((aop->fPos[0] != 0.0f) || (aop->fPos[1] != 0.0f) || (aop->fPos[2] != 0.0f))) {
 				float dir[3] = { aop->fPos[0] - g.p->fCameraPosition[0], aop->fPos[1] - g.p->fCameraPosition[1], aop->fPos[2] - g.p->fCameraPosition[2] };
@@ -1082,7 +1141,7 @@ bool AudioOutput::mix(void *outbuff, unsigned int nsamp) {
 				}
 				for (unsigned int s=0;s<nchan;++s) {
 					const float dot = bSpeakerPositional[s] ? dir[0] * speaker[s*3+0] + dir[1] * speaker[s*3+1] + dir[2] * speaker[s*3+2] : 1.0f;
-					const float str = svol[s] * calcGain(dot, len);
+					const float str = svol[s] * calcGain(dot, len) * volumeAdjustment;
 					float * RESTRICT o = output + s;
 					const float old = (aop->pfVolume[s] >= 0.0) ? aop->pfVolume[s] : str;
 					const float inc = (str - old) / static_cast<float>(nsamp);
@@ -1096,7 +1155,7 @@ bool AudioOutput::mix(void *outbuff, unsigned int nsamp) {
 				}
 			} else {
 				for (unsigned int s=0;s<nchan;++s) {
-					const float str = svol[s];
+					const float str = svol[s] * volumeAdjustment;
 					float * RESTRICT o = output + s;
 					for (unsigned int i=0;i<nsamp;++i)
 						o[i*nchan] += pfBuffer[i] * str;
@@ -1104,6 +1163,10 @@ bool AudioOutput::mix(void *outbuff, unsigned int nsamp) {
 			}
 		}
 
+		if (recorder && recorder->getMixDown()) {
+			recorder->addBuffer(NULL, recbuff, nsamp);
+		}
+
 		// Clip
 		if (eSampleFormat == SampleFloat)
 			for (unsigned int i=0;i<nsamp*iChannels;i++)
@@ -1120,3 +1183,11 @@ bool AudioOutput::mix(void *outbuff, unsigned int nsamp) {
 
 	return (! qlMix.isEmpty());
 }
+
+bool AudioOutput::isAlive() const {
+	return isRunning();
+}
+
+unsigned int AudioOutput::getMixerFreq() const {
+	return iMixerFreq;
+}
